/*
 * sim2.c --
 *
 *      Este fichero contiene parte del simulador
 *	para el DLXV.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <bwcc.h>
#include <alloc.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "tcl.h"
#include "asm.h"
#include "dlx.h"
#include "sym.h"
#include "sim.h"

static char * far operationNames[] = {
    "", "ADD", "ADDI", "ADDU", "ADDUI", "AND", "ANDI", "BEQZ", "BFPF",
    "BFPT", "BNEZ", "J", "JAL", "JALR", "JR", "LB", "LBU", "LD", "LF",
    "LH", "LHI", "LHU", "LW", "MOVD", "MOVF", "MOVFP2I", "MOVI2FP",
    "MOVI2S", "MOVS2I", "OR", "ORI", "RFE", "SB", "SD", "SEQ",  "SEQI",
    "SEQU", "SEQUI",
    "SF", "SGE", "SGEI", "SGEU", "SGEUI", "SGT", "SGTI", "SGTU", "SGTUI",
    "SH", "SLE", "SLEI", "SLEU", "SLEUI", "SLL",
    "SLLI", "SLT", "SLTI", "SLTU", "SLTUI", "SNE", "SNEI", "SNEU", "SNEUI",
    "SRA", "SRAI", "SRL",
    "SRLI", "SUB", "SUBI", "SUBU", "SUBUI", "SW", "TRAP", "XOR", "XORI",
    "NOP", "", "", "","", "",

    /* Debe haber al menos un string vacio entre la ultima operacion entera
     * y la primera de las operaciones en coma flotante. */

    "ADDD", "ADDF", "CVTD2F", "CVTD2I", "CVTF2D", "CVTF2I", "CVTI2D",
    "CVTI2F", "DIV", "DIVD", "DIVF", "DIVU", "EQD", "EQF", "GED", "GEF",
    "GTD", "GTF", "LED", "LEF", "LTD", "LTF", "MULT", "MULTD", "MULTF",
    "MULTU", "NED", "NEF", "SUBD", "SUBF", "",

    /* Separacion entre operaciones FP y vectoriales */

    "ADDSV", "ADDV", "CVI", "CVM", "DIVSV", "DIVV", "DIVVS",
    "LV", "LVI", "LVWS", "MOVF2S", "MOVS2F", "MULTSV", "MULTV",
    "POP", "SEQSV", "SEQV", "SGESV", "SGEV", "SGTSV", "SGTV",
    "SLESV", "SLEV", "SLTSV", "SLTV", "SNESV", "SNEV", "SUBSV",
    "SUBV", "SUBVS", "SV", "SVI", "SVWS",   ""
    };

extern char far salidas[199];
extern Init_Handle_Trap();
/*
 *----------------------------------------------------------------------
 * statsReset --
 *
 *      Este procedimiento inicializa las varias estadisticas que son
 *	guardadas dinamicamente durante la ejecucion del codigo del DLXV.
 *
 * Resultados:
 *      Ninguno
 *
 * Efectos Laterales:
 *      Cambia dichas variables globales.
 *
 *----------------------------------------------------------------------
 */

statsReset(machPtr)
     DLXV *machPtr;                   /* Descripcion de la maquina. */
{
    int i;

    machPtr->loadStalls = 0;
    machPtr->FPstalls = 0;
    machPtr->Vstalls = 0;
    machPtr->Terminar = 0;
    machPtr->branchYes = 0;
    machPtr->branchNo = 0;
    for (i = 0; i <= OP_LAST; i++)
	machPtr->operationCount[i] = 0;

    return 0;
}


/*
 *----------------------------------------------------------------------
 * regsReset --
 *
 *      Este procedimiento inicializa los registros del DLXV.
 *
 * Resultados:
 *      Ninguno
 *
 * Efectos Laterales:
 *      Cambia dichas variables globales.
 *
 *----------------------------------------------------------------------
 */

regsReset(machPtr)
     DLXV *machPtr;                   /* Descripcion de la maquina. */
{
    int i,j;

    for (i = 0; i < TOTAL_REGS; i++) {
	machPtr->regs[(int)i] = 0;
    }
    for (i = 0; i < NUM_V_REGS; i++)
    { for (j = 0; j < (MVL * 2); j++) {
      machPtr->vecregs[(int)i][(int)j] = 0;
      }
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_Create --
 *
 *      Crea la maquina DLXV.
 *
 * Resultados:
 *      El valor devuelto es un puntero a la descripcion de la maquina DLXV.
 *
 * Efectos Laterales:
 *      La estructura DLXV es alojada e inicializada.
 *
 *----------------------------------------------------------------------
 */

DLXV *
Sim_Create(memSize, au, al, mu, ml, du, dl, vau, val, vmu, vml, vdu, vdl, vlu,
	   vll, vgu, vgl, banc, puertos_l, puertos_e)
    long int memSize;           /* Numero de bytes de memoria a ser alojados
				 * en la maquina. */
    long int au, mu, du;        /* latencias para suma, multiplicacion,
				 * y division en coma flotante. */
    long int al, ml, dl;        /* Nmero de unidades de suma, multiplicacion y
				 * division en coma flotante. */
    long int vau, vmu, vdu, vlu, vgu;	/* numero de unidades vectoriales. */
    long int val, vml, vdl, vll, vgl;	/* Penalizaciones para las unidades vectoriales. */
    long int banc;			/* Numero de bancos de memoria */
    int puertos_l, puertos_e; /* Numero de puertos de lectura y escritura
					de los registros vectoriales */
{
    register DLXV *machPtr;
    register MemWord huge *wordPtr;
    long int i, j;
/*    HANDLE hMem; */

    machPtr = (DLXV far *) farcalloc(1, sizeof(DLXV));
    machPtr->numWords = (memSize+3) & ~0x3;
    machPtr->numChars = machPtr->numWords * 4;
    machPtr->memPtr = (MemWord huge *)
	    farcalloc(1, (unsigned long int) (sizeof(MemWord) * machPtr->numWords));
    if (machPtr->memPtr == NULL) {farfree(machPtr);
				  return(NULL);
				 }

    for (i = machPtr->numWords, wordPtr = machPtr->memPtr; i > 0;
	    i--, wordPtr++) {
	wordPtr->value = 0;
	wordPtr->opCode = OP_NOT_COMPILED;
	wordPtr->stopList = NULL;
    }
    machPtr->memScratch = (char huge *) farmalloc (machPtr->numChars);
    if (machPtr->memScratch==NULL) {farfree(machPtr->memPtr);
				    farfree(machPtr);
				    return(NULL);}
    machPtr->endScratch = machPtr->memScratch + machPtr->numChars;
    for (i = 0; i < NUM_GPRS; i++) {
	machPtr->regs[(int)i] = 0;
    }
    for (i = 0; i < NUM_V_REGS; i++)
    { for (j = 0; j < (MVL * 2); j++) {
      machPtr->vecregs[(int)i][(int)j] = 0;
      }
    }

    for(i=0; i < NUM_V_REGS; i++){
	machPtr->rports[(int)i] = 0;
	machPtr->wports[(int)i] = 0;
	machPtr->waitS[(int)i] = 0;
    }

    machPtr->read_ports = puertos_l;
    machPtr->write_ports = puertos_e;
    machPtr->bancos = banc;
    machPtr->regs[(int)PC_REG] = 0;
    machPtr->regs[(int)NEXT_PC_REG] = 1;
    machPtr->regs[(int)(NEXT_PC_REG+1)] = 0;
    machPtr->regs[(int)(NEXT_PC_REG+2)] = 0;
    machPtr->regs[(int)(NEXT_PC_REG+3)] = 0;
    machPtr->badPC = 0;
    machPtr->addrErrNum = 0;
    machPtr->loadReg1 = 0;
    machPtr->loadValue1 = 0;
    machPtr->loadReg2 = 0;
    machPtr->loadValue2 = 0;
    machPtr->insCount = 0;
    machPtr->firstIns = 0;
    machPtr->branchSerial = -1;
    machPtr->branchPC = 0;
    machPtr->flags = 0;
    machPtr->Terminar = 0;
    machPtr->stopNum = 1;
    machPtr->num_add_units = au;
    machPtr->num_mul_units = mu;
    machPtr->num_div_units = du;
    machPtr->fp_add_latency = al;
    machPtr->fp_mul_latency = ml;
    machPtr->fp_div_latency = dl;
    machPtr->num_addv_units = vau;
    machPtr->num_mulv_units = vmu;
    machPtr->num_divv_units = vdu;
    machPtr->num_loadv_units = vlu;
    machPtr->num_logic_units = vgu;
    machPtr->v_add_latency = val;
    machPtr->v_mul_latency = vml;
    machPtr->v_div_latency = vdl;
    machPtr->v_load_latency = vll;
    machPtr->v_logic_latency = vgl;
    machPtr->stopList = NULL;
    machPtr->callBackList = NULL;
    Hash_InitTable(&machPtr->symbols, cero, HASH_STRING_KEYS);
    Io_Init(machPtr);
    Cop0_Init(machPtr);
    Init_Handle_Trap(machPtr);

    /* Inicializa todas las estructuras relacionadas con las diferentes
     * unidades de proceso. */
    machPtr->FPstatusReg = 0;
    for (i = 0; i < machPtr->num_add_units; i++) machPtr->fp_add_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_div_units; i++) machPtr->fp_div_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_mul_units; i++) machPtr->fp_mul_units[(int)i] = 0;

    for (i = 0; i < machPtr->num_addv_units; i++) machPtr->v_add_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_divv_units; i++) machPtr->v_div_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_mulv_units; i++) machPtr->v_mul_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_loadv_units; i++) machPtr->v_load_units[(int)i] = 0;
    for (i = 0; i < machPtr->num_logic_units; i++) machPtr->v_logic_units[(int)i] = 0;

    machPtr->fp_units[(int)FP_ADD] = machPtr->fp_add_units;
    machPtr->fp_units[(int)FP_DIV] = machPtr->fp_div_units;
    machPtr->fp_units[(int)FP_MUL] = machPtr->fp_mul_units;

    machPtr->fp_units[(int)V_ADD] = machPtr->v_add_units;
    machPtr->fp_units[(int)V_DIV] = machPtr->v_div_units;
    machPtr->fp_units[(int)V_MUL] = machPtr->v_mul_units;
    machPtr->fp_units[(int)V_LOAD] = machPtr->v_load_units;
    machPtr->fp_units[(int)V_LOGIC] = machPtr->v_logic_units;

    for (i = 0; i < (32 + NUM_V_REGS + 1); i++) {
    machPtr->waiting_FPRs[(int)i] = machPtr->waitingE[(int)i] = 0;}
    machPtr->cycleCount = 0;
    machPtr->FPopsList = NULL;
    machPtr->checkFP = 0;
    machPtr->refTraceFile = NULL;
    machPtr->result = machPtr->resultSpace;
    machPtr->dynamic = 0;
    machPtr->errorLine = 0;
    machPtr->resultSpace[0] = 0;
/* Inicializa los contadores */
    statsReset(machPtr);

    Sym_AddSymbol(machPtr, "", "memSize", memSize * 4, SYM_GLOBAL);

    return machPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_DumpStats --
 *
 *      Este procedimiento es invocado para procesar el comando
 *	"Estadisticas".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no se produce error.
 *	TCL_ERROR si se produce algun error.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Sim_DumpStats(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    long int i, j, intCount, floatCount, vectCount;
    FILE *nombref;
    long int doStalls = 0, doOpCount = 0, doFPpending = 0,
	doBranch = 0, doHW = 0, doReset = 0;
    char *opName = *argumentos;
    long int exec_time, instr_count;

    /* Por defecto se muestran todas las estadsticas */



    if (numarg == 1) {
	doStalls = doOpCount = doFPpending = doBranch = doHW = 1;
    } else while (++argumentos, --numarg) {
	if (!strcmp(*argumentos, "r") || !strcmp(*argumentos, "R")) {
	    doReset = 1;
	} else if (!strcmp(*argumentos, "s") || !strcmp(*argumentos, "S")) {
	    doStalls = 1;
	} else if (!strcmp(*argumentos, "o") || !strcmp(*argumentos, "O")) {
	    doOpCount = 1;
	} else if (!strcmp(*argumentos, "p") || !strcmp(*argumentos, "P")) {
	    doFPpending = 1;
	} else if (!strcmp(*argumentos, "b") || !strcmp(*argumentos, "B")) {
	    doBranch = 1;
	} else if (!strcmp(*argumentos, "h") || !strcmp(*argumentos, "H")) {
	    doHW = 1;
	} else if (!strcmp(*argumentos, "a") || !strcmp(*argumentos, "A")) {
		/* all prints everything, but does not reset */
	    doStalls = doOpCount = doFPpending = doBranch = doHW = 1;
	} else {
	    wsprintf(salidas,
		"Illegal option:  must be \"%.50s\" [reset] [stalls] [opcount] [pending] [all]", opName);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    }
    if (!doReset) { nombref=fopen("Estad.tmp", "wt");
		    machPtr->result[0] = '\1';
		  }
    if (doHW) {
      fprintf(nombref,"\n---------\n");
      fprintf(nombref,"HARDWARE:\n");
      fprintf(nombref,"---------\n");
      fprintf(nombref,"Memory: %ld bytes, distributed in %ld banks.\n", machPtr->numWords * 4, machPtr->bancos);
      fprintf(nombref,"\nScalar floating point Units\n");
      fprintf(nombref,"%2ld unit(s) suma,\t\tlatency = %2ld cycles\n",
	     machPtr->num_add_units, machPtr->fp_add_latency);
      fprintf(nombref,"%2ld unit(s) division,\t\tlatency = %2ld cycles\n",
	     machPtr->num_div_units, machPtr->fp_div_latency);
      fprintf(nombref,"%2ld unit(s) multiplicacion,\tlatency = %2ld cycles\n",
	     machPtr->num_mul_units, machPtr->fp_mul_latency);

      fprintf(nombref,"\nVector Units\n");
      fprintf(nombref,"%2ld unit(s) suma,\t\tlatency = %2ld cycles\n",
	     machPtr->num_addv_units, machPtr->v_add_latency);
      fprintf(nombref,"%2ld unit(s) division,\t\tlatency = %2ld cycles\n",
	     machPtr->num_divv_units, machPtr->v_div_latency);
      fprintf(nombref,"%2ld unit(s) multiplicacion,\tlatency = %2ld cycles\n",
	     machPtr->num_mulv_units, machPtr->v_mul_latency);
      fprintf(nombref,"%2ld unit(s) load/store,\t\tlatency = %2ld cycles\n",
	     machPtr->num_loadv_units, machPtr->v_load_latency);
      fprintf(nombref,"%2ld unit(s) logical(s),\t\tlatency = %2ld cycles\n",
	     machPtr->num_logic_units, machPtr->v_logic_latency);
      fprintf(nombref,"\n%d/%d read/write ports in vector registers\n",
	     machPtr->read_ports, machPtr->write_ports);

    }

    if (doStalls) {
	fprintf(nombref,"\n--------------------\n");
	fprintf(nombref,"    CYCLES COUNT:\n");
        fprintf(nombref,"--------------------\n");
	fprintf(nombref,"Cycles in entered instructions:\n\n");
	fprintf(nombref,"  Enter. instr. load/store        = %ld\n", machPtr->loadStalls);
	fprintf(nombref,"  Enter. instr. floating point    = %ld\n", machPtr->FPstalls);
	fprintf(nombref,"  Enter. instr. vectors           = %ld\n", machPtr->Vstalls);
	for (intCount = 0, i = 1; *operationNames[(int)i]; i++) {
	    intCount += machPtr->operationCount[(int)i];
	}
	while (!*operationNames[(int)i])
	    i++;
	for (floatCount = j = 0; *operationNames[(int)i]; i++, j++) {
	    floatCount += machPtr->operationCount[(int)i];
	}
	while (!*operationNames[(int)i])
	    i++;
	for (vectCount = j = 0; *operationNames[(int)i]; i++, j++) {
	    vectCount += machPtr->operationCount[(int)i];
	}
        instr_count = intCount + floatCount + vectCount;
	exec_time = intCount + floatCount + vectCount +
	       machPtr->loadStalls + machPtr->FPstalls + machPtr->Terminar + machPtr->Vstalls;
	fprintf(nombref,"\nProgram execution time = %ld cycles\n", exec_time);
	fprintf(nombref,"\nNumber of executed instructions = %l\n", instr_count);
	if (instr_count == 0) {instr_count = 1;}; // evitar division por cero
	fprintf(nombref,"\nCPI = %2.2f cycles\n", (float) exec_time/(float) instr_count);
    }

    if (doBranch) {
      long int total = machPtr->branchYes + machPtr->branchNo;
      fprintf(nombref,"\n------------------------\n");
      fprintf(nombref,"   JUMP INSTRUCTIONS:\n");
      fprintf(nombref,"------------------------\n");
      if (!total) fprintf (nombref,"No jump instructions executed.\n");
      else
	fprintf (nombref,"\nTotal %, executed %ld (%.2lf%%), not executed %ld (%.2lf%%)\n",
		total, machPtr->branchYes, 100.0 * machPtr->branchYes / total,
		machPtr->branchNo, 100.0 * machPtr->branchNo / total);
    }

    if (doFPpending) {
	/* Imprime el estado de las unidades de proceso */
	fprintf(nombref,"\n------------------------\n");
	fprintf(nombref," UNFINISHED INSTRUCTIONS:\n");
        fprintf(nombref,"------------------------\n");
	if (machPtr->FPopsList == NULL) fprintf (nombref,"none.\n");
	else {
	  FPop *opPtr;
	  for (opPtr = machPtr->FPopsList; opPtr != NULL; opPtr = opPtr->nextPtr) {
	    switch (opPtr->type) {
	    case FP_ADD: fprintf (nombref,"adder           "); break;
	    case FP_SUB: fprintf (nombref,"adder (sub)     "); break;
	    case INT_UDIV:
	    case INT_DIV: fprintf(nombref,"divider (int)   "); break;
	    case FP_DIV: fprintf (nombref,"divider         "); break;
	    case INT_MUL:
	    case INT_UMUL:fprintf(nombref,"multiplier (int)"); break;
	    case FP_MUL: fprintf (nombref,"multiplier      "); break;
	    case V_ADD: fprintf (nombref, "Vadder	   "); break;
	    case V_SUB: fprintf (nombref, "Vadder (sub)	   "); break;
	    case V_DIV: fprintf (nombref, "Vdivider        "); break;
	    case V_MUL: fprintf (nombref, "Vmultiplier     "); break;
	    case V_LOAD: fprintf (nombref,"Vload/store     "); break;
	    case V_LOGIC:
	      fprintf (nombref,"Vlogic          ");
	      fprintf (nombref," #%1ld :  completed %2ld cycle(s);", opPtr->unit,
	       opPtr->ready - machPtr->cycleCount);
	      fprintf (nombref,"  %lf ==> VM\n",
	       *((double *)&(opPtr->result[0]))); 
	      break;
	    }
            if (opPtr->type != V_LOGIC) {
	      fprintf (nombref," #%1ld :  completed %2ld cycle(s);", opPtr->unit,
		    opPtr->ready - machPtr->cycleCount);}
	    switch (opPtr->resultType) {
	    case FP_INT_OP:
	      fprintf(nombref,"  0x%08lx ==> F%1ld\n", *((long int *)&(opPtr->result[0])), opPtr->dest);
	      break;
	    case FP_SINGLE_FP_OP:
	      fprintf(nombref,"  %f ==> F%1ld\n", *((float *)&(opPtr->result[0])), opPtr->dest);
	      break;
	    case FP_DOUBLE_FP_OP:
	      fprintf (nombref,"  %lf ==> F%1ld:F%1ld\n",
		      *((double *)&(opPtr->result[0])), opPtr->dest, opPtr->dest + 1);
	      break;
	    case FP_VECT_OP:
	    case FP_SV_OP:
	    case FP_VS_OP:
	    case FP_MASK_OP:
	    case FP_LOAD_OP:
	    case FP_LSTD_OP:
	    case FP_LINDEX_OP:
	      fprintf (nombref,"  %lf ==> V%1ld\n",
		      *((double *)&(opPtr->result[0])), (opPtr->dest - 32));
	      break;
	    case FP_STORE_OP:
	      fprintf (nombref,"  %lf ==> dir: %ld\n",
		      *((double *)&(opPtr->result[0])), (opPtr->dest));
	      break;
	    case FP_SSTD_OP:
	    case FP_SINDEX_OP:
	      fprintf (nombref,"  %lf ==> dir: %ld\n",
		      *((double *)&(opPtr->result[0])), (opPtr->direccion));
	      break;
	    }
	  }
	}
	fprintf(nombref,"\nVector register:   0   1   2   3   4   5   6   7\n");
	fprintf(nombref," Read ports:  %4d%4d%4d%4d%4d%4d%4d%4d\n",machPtr->rports[0],
		machPtr->rports[1],machPtr->rports[2],machPtr->rports[3],machPtr->rports[4],
		machPtr->rports[5],machPtr->rports[6],machPtr->rports[7]);
	fprintf(nombref," Write ports:%4d%4d%4d%4d%4d%4d%4d%4d\n",machPtr->wports[0],
		machPtr->wports[1],machPtr->wports[2],machPtr->wports[3],machPtr->wports[4],
		machPtr->wports[5],machPtr->wports[6],machPtr->wports[7]);
    }

    if (doOpCount) {
	fprintf(nombref,"\n-------------------------\n");
	fprintf(nombref,"   INSTRUCTIONS EXECUTED:\n");
	fprintf(nombref,"-------------------------\n");
	fprintf(nombref,"\t\tINSTRUCTIONS ENTERED\n");
	fprintf(nombref,"\t\t=====================\n");
	for (intCount = 0, i = 1; *operationNames[(int)i]; i++) {
	    if (i % 4 == 1)
		fprintf(nombref,"\n");
	    fprintf(nombref,"%s\t%ld\t", operationNames[(int)i], machPtr->operationCount[(int)i]);
	    intCount += machPtr->operationCount[(int)i];
	}
	fprintf(nombref,"\nTotal entered instructions = %ld\n\n", intCount);

	/* Busca la primera operacion en coma flotante */
	while (!*operationNames[(int)i])
	    i++;

	fprintf(nombref,"\t\t FLOATING POINT INSTRUCTIONS\n");
	fprintf(nombref,"\t\t==============================\n");
	for (floatCount = j = 0; *operationNames[(int)i]; i++, j++) {
	    if (j % 4 == 0)
		fprintf(nombref,"\n");
	    fprintf(nombref,"%s\t%ld\t", operationNames[(int)i], machPtr->operationCount[(int)i]);
	    floatCount += machPtr->operationCount[(int)i];
	}
	fprintf(nombref,"\nTotal number of FP instructions = %ld\n\n", floatCount);

	/* Busca la primera operacion vectorial */
	while (!*operationNames[(int)i])
	    i++;

	fprintf(nombref,"\t\tVECTOR INSTRUCTIONS\n");
	fprintf(nombref,"\t\t===================\n");
	for (vectCount = j = 0; *operationNames[(int)i]; i++, j++) {
	    if (j % 4 == 0)
		fprintf(nombref,"\n");
	    fprintf(nombref,"%s\t%ld\t", operationNames[(int)i], machPtr->operationCount[(int)i]);
	    vectCount += machPtr->operationCount[(int)i];
	}
	instr_count = intCount + floatCount + vectCount;
	fprintf(nombref,"\nTotal number of vector instructions = %ld\n", vectCount);
	fprintf(nombref,"\nTotal number of instructions     = %ld\n", intCount + floatCount + vectCount);
	fprintf(nombref,"\nTotal number of cycles          = %ld\n", exec_time);
	if (instr_count == 0) {instr_count = 1;}; // evitar division por cero 
	fprintf(nombref,"\nCPI = %2.2f cycles\n", (float) exec_time/(float) instr_count);
    }

    if (doReset) {
	statsReset(machPtr);
    }

    if (!doReset) { fclose(nombref);}

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_GoCmd --
 *
 *      Este procedimiento es invocado para procesar el comando
 *	"Ejecutar".
 *      Ver la documentacion de usuario para mas detalles
 *
 * Resultados:
 *      TCL_OK si no se producen errores.
 *	TCL_ERROR si se produce algun error.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Sim_GoCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{

    if (numarg > 2) {
	wsprintf(salidas,
		"Too many arguments:  must be \"%.50s\" [direccin]", argumentos[0]);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }

    if (numarg == 2) {
	char *end;
	long int newPc;

	if (Sym_EvalExpr(machPtr, (char far *) NULL, argumentos[1], cero, &newPc, &end)
		!= TCL_OK) {
	    return TCL_ERROR;

	  }
	if ((*end != 0) || (newPc & 0x3)) {
	    wsprintf(salidas,
		    "\"%.50s\" Invalid beginning position", argumentos[1]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	machPtr->regs[(int)PC_REG] = ADDR_TO_INDEX(newPc);
	machPtr->regs[(int)NEXT_PC_REG] = machPtr->regs[(int)PC_REG] + 1;
	machPtr->flags = 0;
	machPtr->badPC = 0;
    }

    return Simulate(machPtr, cero, numarg, argumentos);
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_StepCmd --
 *
 *      Este procedimiento es invocado para procesar el comando "Paso
 *	a Paso".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores.
 *	TCL_ERROR si hay errores.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Sim_StepCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{

    if (numarg > 2) {
	wsprintf(salidas,
		"Too many arguments:  must be \"%.50s\" [direction]", argumentos[0]);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }

    if (numarg == 2) {
	char *end;
	long int newPc;

	if (Sym_EvalExpr(machPtr, (char far *) NULL, argumentos[1], cero, &newPc, &end)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	if ((*end != 0) || (newPc & 0x3)) {
	    wsprintf(salidas,
		    "\"%.50s\" Invalid position", argumentos[1]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	machPtr->regs[(int)PC_REG] = ADDR_TO_INDEX(newPc);
	machPtr->regs[(int)NEXT_PC_REG] = machPtr->regs[(int)PC_REG] + 1;
	machPtr->flags = 0;
	machPtr->badPC = 0;
    }

    return Simulate(machPtr, uno, numarg, argumentos);
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_CallBack --
 *
 *      Permite que un determinado procedimiento pueda ser invocado despues
 *	de que un determinado numero de instrucciones hayan sido simuladas.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Despues de numIns instrucciones ejecutadas se invoca de la siguiente
 *	manera:
 *
 *      void
 *      proc(clientData, machPtr)
 *          ClientData clientData;
 *          DLXV *machPtr;
 *      {
 *      }
 *
 *      Los argumentos clientData y machPtr seran los mismos que los pasados
 *	a este procedimiento.
 *----------------------------------------------------------------------
 */

void
Sim_CallBack(machPtr, numIns, proc, clientData)
    DLXV *machPtr;               /* Descripcion de la maquina. */
    long int numIns;            /* Llamar al proc despues de que se hayan
				 * ejecutado estas instrucciones. */
    void (*proc)();             /* Procedimiento a llamar. */
    ClientData clientData;      /* Valor arbitrario de una unica palabra
				 * que es pasado a proc. */
{
    register CallBack *cbPtr;

    cbPtr = (CallBack far *) farcalloc(1, sizeof(CallBack));
    cbPtr->serialNum = machPtr->insCount + numIns;
    cbPtr->proc = proc;
    cbPtr->clientData = clientData;
    if ((machPtr->callBackList == NULL) ||
	    (cbPtr->serialNum < machPtr->callBackList->serialNum)) {
	cbPtr->nextPtr = machPtr->callBackList;
	machPtr->callBackList = cbPtr;
    } else {
	register CallBack *cbPtr2;

	for (cbPtr2 = machPtr->callBackList; cbPtr2->nextPtr != NULL;
		cbPtr2 = cbPtr2->nextPtr) {
	    if (cbPtr->serialNum < cbPtr2->nextPtr->serialNum) {
		break;
	    }
	}
	cbPtr->nextPtr = cbPtr2->nextPtr;
	cbPtr2->nextPtr = cbPtr;
    }
    if (cbPtr->serialNum < checkNum) {
	checkNum = cbPtr->serialNum;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_GetPC --
 *
 *      Este procedimiento calcula el contador de programa actual para
 *      machPtr.
 *
 * Resultados:
 *      El valor devuelto es el actual contador de programa de la maquina.
 *	Esto es un poco complicado de calcular ya que el PC es almacenado
 *	como un indice, y puede haber un valor no alineado en el mismo.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

unsigned long int
Sim_GetPC(machPtr)
    register DLXV *machPtr;     /* Maquina de la cual nos interesa el PC. */
{
    if ((machPtr->badPC != 0) && (machPtr->insCount >= machPtr->addrErrNum)) {
	return machPtr->badPC;
    }
    return INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_Stop --
 *
 *      Programa la parada de la ejecucion de la maquina despues de la
 *	instruccion actual.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      La maquina parara su ejecucion (si estaba ejecutando)
 *
 *----------------------------------------------------------------------
 */

void
Sim_Stop(machPtr)
    DLXV *machPtr;                       /* Maquina a parar. */
{
    machPtr->flags |= STOP_REQUESTED;
    checkNum = machPtr->insCount + 1;
}

/*
 *----------------------------------------------------------------------
 *
 * Sim_TraceCmd --
 *
 *      Este comando activa o desactiva la salida del simulador para el
 *	programa Dinero III.
 *
 * Resultados:
 *      TCL_OK si no hay errores.
 *	TCL_ERROR si hay errores.
 *
 * Efectos Laterales:
 *      Ver documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */
long int
Sim_TraceCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{

	if( numarg == 2 && !strcmp( argumentos[1], "off" ) ) {
		if( !machPtr->refTraceFile ) {
			BWCCMessageBox(GetFocus(),"Output for program \"Dinero III\" not established", "Error", MB_OK | MB_ICONEXCLAMATION);
			machPtr->result[0] = '\0';
			return TCL_ERROR;
		}
		fclose( machPtr->refTraceFile );
		machPtr->refTraceFile = NULL;
		return TCL_OK;
	}
	if( numarg == 3 && !strcmp( argumentos[1], "on" ) ) {
		if( machPtr->refTraceFile ) {
			BWCCMessageBox(GetFocus(),
			  "Output for program \"Dinero III\" is activated -- First deactivate", "Error", MB_OK | MB_ICONEXCLAMATION);
			  machPtr->result[0] = '\0';
			return TCL_ERROR;
		}
		if( !(machPtr->refTraceFile = fopen(argumentos[2], "wt" )) ) {
			wsprintf( salidas, "cannot open \"%s\": %s",
			  argumentos[2], strerror(errno));
			BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
			machPtr->result[0] = '\0';
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	BWCCMessageBox(GetFocus(), "Incorrect option: must be <Filename> 'On' or 'Off'.", "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0]='\0';
	return TCL_ERROR;
}