/*
 * stop.c --
 *
 *	Este procedimiento proporciona el codigo que manipula los
 *	espias y los stops para el simulador.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <bwcc.h>
#include <alloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "tcl.h"
#include "asm.h"
#include "dlx.h"
#include "sym.h"

long int
Elegir(machPtr, cm, numarg, argumentos)
DLXV	*machPtr;
char 	cm[25];
int	numarg;
char 	argumentos[argus][largus];

{
long int resul = TCL_OK;
	if ((strcmp(cm,"asm"))==0) {resul=Asm_AsmCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"fget"))==0) { resul=Gp_FGetCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"fput"))==0) { resul=Gp_FPutCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"get"))==0) { resul=Gp_GetCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"go"))==0) { resul=Sim_GoCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"load"))==0) { resul=Asm_LoadCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"put"))==0) { resul=Gp_PutCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"step"))==0) { resul=Sim_StepCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"stop"))==0) { resul=Stop_StopCmd(machPtr, numarg, argumentos);goto result;}
	if ((strcmp(cm,"trace"))==0) { resul=Sim_TraceCmd(machPtr, numarg, argumentos);goto result;}
	BWCCMessageBox(GetFocus(), "Unknown stop command", "Error", MB_OK | MB_ICONEXCLAMATION);
result: return resul;
}

/*
 *----------------------------------------------------------------------
 *
 * Stop_StopCmd --
 *
 *	Este procedimiento es invocado para procesar el comando "stop".
 *	Ver documentacion de usuario para mas detalles.
 *
 * Resultados:
 *	TCL_OK si no hay errores.
 *	TCL_ERROR si hay errores.
 *
 * Efectos Laterales:
 *	Ver documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Stop_StopCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;			/* Descripcion de la maquina. */
    int numarg;				/* Numero de argumentos. */
    char argumentos[argus][largus];	/* Cadena de argumentos. */
{
    long int length, indj, tamao;
    register Stop *stopPtr;
    Stop *prevPtr;
    MemWord huge *wordPtr;
    extern char far salidas[199];
    FILE *nombref;

    if (numarg == 1) {
	Sim_Stop(machPtr);
	return TCL_OK;
    }

    length = _fstrlen(argumentos[1]);

    /*
     * Opcion "En" :  crea un nuevo stop.
     */

    if (_fstrncmp(argumentos[1], "at", (unsigned int)length) == 0) {
	unsigned long int address;
	long int result;
	char *end;

	if ((numarg != 3) && (numarg != 4)) {
	    wsprintf(salidas,
		    "Incorrect # arguments:  must be \"%.50s at posicion [command]\"",
		    argumentos[0]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	result = Sym_EvalExpr(machPtr, (char far *) NULL, argumentos[2], cero,
		(long int *) &address, &end);
	if (result != TCL_OK) {
	    return result;
	}
	if (*end != 0) {
	    wsprintf(salidas,
		    "Wrong position \"%.50s\" in command \"%.50s\" ",
		    argumentos[2], argumentos[0]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	address = ADDR_TO_INDEX(address);
	if (address >= machPtr->numWords) {
	    wsprintf(salidas,
		    "Posicion %.50s Inexisting in memory",
		    argumentos[2]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	wordPtr = &machPtr->memPtr[address];
	for (stopPtr = wordPtr->stopList; stopPtr != NULL;
		stopPtr = stopPtr->nextPtr) {
	    if (stopPtr->command == NULL) {
		wsprintf(salidas, "Stop at %.50s",
			argumentos[2]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_OK;
	    }
	    prevPtr = stopPtr;
	}
	stopPtr = (Stop far *) farcalloc(1, sizeof(Stop));
	if (numarg == 3) {
	    stopPtr->command = (char far *) NULL;
	} else {
	    tamao = _fstrlen(argumentos[3]);
	    stopPtr->command = farcalloc(1, (unsigned long int) (tamao+1));
	    for (indj=0;indj<tamao;indj++)
		{ stopPtr->command[(unsigned int)indj] = argumentos[3][(unsigned int)indj];
		}
	    stopPtr->command[(unsigned int)indj] = '\x0';
	}
	stopPtr->number = machPtr->stopNum;
	machPtr->stopNum += 1;
	stopPtr->address = INDEX_TO_ADDR(address);
	stopPtr->nextPtr = NULL;
	stopPtr->overallPtr = NULL;
	if (wordPtr->stopList == NULL) {
	    wordPtr->stopList = stopPtr;
	} else {
	    prevPtr->nextPtr = stopPtr;
	}
	if (machPtr->stopList == NULL) {
	    machPtr->stopList = stopPtr;
	} else {
	    for (prevPtr = machPtr->stopList; prevPtr->overallPtr != NULL;
		    prevPtr = prevPtr->overallPtr) {
		/* Empty loop body. */
	    }
	    prevPtr->overallPtr = stopPtr;
	}
	return TCL_OK;
    }

    /*
     * Opcion "Borrar" :  elimina un stop.
     */

    if (_fstrncmp(argumentos[1], "delete", (unsigned int)length) == 0) {
	long int number, i;
	char *end;

	if (numarg < 3) {
	    wsprintf(salidas,
		    "Incorrect # arguments:  must be \"%.50s delete num num ...\"",
		    argumentos[0]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	for (i = 2; i < numarg; i++) {
	    number = strtol(argumentos[(int)i], &end, 0);
	    if (*end != 0) {
		wsprintf(salidas, "Stop number incorrect \"%.50s\"", argumentos[(int)i]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    }
	    for (stopPtr = machPtr->stopList; stopPtr != NULL;
		    stopPtr = stopPtr->overallPtr) {
		if (stopPtr->number == number) {
		    break;
		}
		prevPtr = stopPtr;
	    }
	    if (stopPtr == NULL) {
		wsprintf(salidas, "No stop definition \"%.50s\" ",
			argumentos[(int)i]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    }
	    if (machPtr->stopList == stopPtr) {
		machPtr->stopList = stopPtr->overallPtr;
	    } else {
		prevPtr->overallPtr = stopPtr->overallPtr;
	    }
	    wordPtr = &machPtr->memPtr[ADDR_TO_INDEX(stopPtr->address)];
	    if (wordPtr->stopList == stopPtr) {
		wordPtr->stopList = stopPtr->nextPtr;
	    } else {
		for (prevPtr = wordPtr->stopList; prevPtr != NULL;
			prevPtr = prevPtr->nextPtr) {
		    if (prevPtr->nextPtr == stopPtr) {
			prevPtr->nextPtr = stopPtr->nextPtr;
			break;
		    }
		}
	    }
	    if (stopPtr->command != NULL) {
		farfree(stopPtr->command);
	    }
	    farfree((char far *) stopPtr);
	}
	return TCL_OK;
    }

    /*
     * Opcion "Info" :  imprime informacion acerca de todos los stops
     * actualmente definidos.
     */

    if (_fstrncmp(argumentos[1], "info", (unsigned int)length) == 0) {
	machPtr->result[0] = '\1';
	nombref=fopen("Stopin.tmp","wt");
	if (machPtr->stopList == NULL) {
	    fprintf(nombref, "No stop points assigned.\n");
	    fclose(nombref);
	    return TCL_OK;
	}
	fprintf(nombref, "Number\tPosition\t\t\tAction\n");
	for (stopPtr = machPtr->stopList; stopPtr != NULL;
		stopPtr = stopPtr->overallPtr) {
	    fprintf(nombref, "#%-2ld\t%-20s", stopPtr->number,
		    Sym_GetString(machPtr, stopPtr->address));
	    if (stopPtr->command != NULL) {
		fprintf(nombref, "\t\t%s\n", stopPtr->command);
	    } else {
		fprintf(nombref, "\t\tstop\n");
	    }
	}
	fclose(nombref);
	return TCL_OK;
    }

    wsprintf(salidas,
	    "Option incorrect \"%.50s\" for \"%.50s\":  %s",
	    argumentos[1], argumentos[0], "must be at, delete, or info");
    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * Stop_Execute --
 *
 *	Busca a lo largo de la lista de espias asociados con una palabra
 *	de memoria, ejecutando el comando asociado con cada uno. Si alguno
 *	de los comandos devuelve un resultado que no sea TCL_OK, entonces
 *	salimos inmendiatamente.
 *
 * Resultados:
 *	TCL_OK, si los comandos asociados con el stop terminaron correctamente;
 *	en otro caso el resultado del primer comando que no termino
 *	correctamente.
 *	MachPtr->result contiene tambien un resultado.
 *
 * Efectos Laterales:
 *	Los que tenga el comando que se ejecuta.
 *
 *----------------------------------------------------------------------
 */

long int
Stop_Execute(machPtr, stopPtr, numarg, argumentos)
    DLXV *machPtr;		/* Maquina que se esta ejecutando. */
    register Stop *stopPtr;	/* Primero en la lista de stops. */
    int numarg;
    char argumentos[argus][largus];
{
    long int result = TCL_OK;
    long int stopFlag = 0;
    long int stopNum = -1;
    char com[25];
    long int i, nivel, indi, j;
    /*
     * El siguiente codigo comprueba cada stop para ver si es el que ha
     * causado la actual parada (el primero stop detectado es tratado
     * como el stop que provoco la parada).
     */

    for ( ; stopPtr != NULL; stopPtr = stopPtr->nextPtr) {
	if (machPtr->flags & STOP_REQUESTED) {
	    stopFlag |= STOP_REQUESTED;
	    machPtr->flags &= ~STOP_REQUESTED;
	}
	if (stopPtr->command == NULL) {
	    Sim_Stop(machPtr);
	} else {
	i = 0;
	while (isspace(stopPtr->command[(int)i])) { i++;}
	nivel = 0;
	indi = 0;
	j=i;
	while (1)
	{ if (isspace(stopPtr->command[(int)j])) { if (nivel == 0) {com[(int)indi] = '\0';
							       argumentos[(int)nivel][(int)indi]= '\0';
							       }
				 else {argumentos[(int)nivel][(int)indi] = '\0'; }
				 nivel++;
				 j++;
				 indi = 0;
				 while (isspace(stopPtr->command[(int)j])) { j++;}
				}
	   else {
		if ((stopPtr->command[(int)j] == '\x0')) {goto fin;}
		   else {
		   if ((stopPtr->command[(int)j] == '{')) { j++;
					  while (isspace(stopPtr->command[(int)i])) { j++;}
					  while (stopPtr->command[(int)j] != '}')
					  { if (nivel == 0)
					    { com[(int)indi] = stopPtr->command[(int)j];}
					    else {
						argumentos[(int)nivel][(int)indi] = stopPtr->command[(int)j];}
					   indi++;
					   j++;
					   }
					  j++;
					 }
		   else {if (nivel == 0) { com[(int)indi] = stopPtr->command[(int)j];
					    argumentos[(int)nivel][(int)indi] = stopPtr->command[(int)j];
					   }
			  else {
			  argumentos[(int)nivel][(int)indi] = stopPtr->command[(int)j];}
			  indi++;
			  j++;
			 }
		   }
		}
	}
fin:    if (nivel==0) {com[(int)indi]='\x0';}
		else {argumentos[(int)nivel][(int)indi]='\x0';}
	numarg=(int)(nivel+1);
	result=Elegir(machPtr, com, numarg, argumentos);
	    if (result != TCL_OK) {
		return result;
	    } else {
		if (*machPtr->result != 0) {
		    MessageBox(GetFocus(), (char far *)machPtr->result, "Stop: Message", MB_OK | MB_ICONINFORMATION);
		}
		Tcl_Return(machPtr, (char far *) NULL, TCL_STATIC);
	    }
	}
	if ((machPtr->flags & STOP_REQUESTED) && (stopNum == -1)) {
	    stopNum = stopPtr->number;
	}
    }
    machPtr->flags |= stopFlag;
    if (stopNum != -1) {
	unsigned long int pcAddr;

	pcAddr = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
	sprintf((char far *)machPtr->result, "stop %ld, pc = %.50s: %.80s", stopNum,
		Sym_GetString(machPtr, pcAddr), Asm_Disassemble(machPtr,
			machPtr->memPtr[machPtr->regs[(int)PC_REG]].value, pcAddr));
    }
    return TCL_OK;
}
