/*
 * trap.c --
 *
 *	Este fichero contiene funciones que manejan varios numeros
 *	de traps, las cuales son usadas para realizar llamadas al
 *	sistema.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <bwcc.h>
#include <dos.h>
#include "tcl.h"
#include <errno.h>
#include "asm.h"
#include "dlx.h"
#include "sym.h"
#include "traps.h"

#define RETURN_VALUE  (machPtr->regs[1])
#define ARG(n)  (machPtr->memPtr[ADDR_TO_INDEX(machPtr->regs[14])+(n)].value)
#define SCRATCH_ADDR(n)  ((char huge *)machPtr->memScratch + (n))

/*
 *----------------------------------------------------------------------
 *
 * Init_Handle_Trap --
 *
 *	Prepara el sistema para manejar las traps.  Esto principalmente
 *	hace copias del stdin, stdout y stderr para que sean usadas por
 *	el programa simulado.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos laterales:
 *      El mapa de descriptores de fichero es inicializado.
 *
 *----------------------------------------------------------------------
 */

Init_Handle_Trap(machPtr)
DLXV *machPtr;
{
  int i;

  for (i = 0; i < FD_SIZE; i++)
    machPtr->fd_map[i] = -1;
  if ((machPtr->fd_map[0] = dup(0)) == -1)
    BWCCMessageBox(GetFocus(),"No se pudo abrir la entrada estndar simulada\n", "Error", MB_OK | MB_ICONEXCLAMATION);
  if ((machPtr->fd_map[1] = dup(1)) == -1)
    BWCCMessageBox(GetFocus(),"No se pudo abrir la salida estndar simulada\n", "Error", MB_OK | MB_ICONEXCLAMATION);
  if ((machPtr->fd_map[2] = dup(2)) == -1)
    BWCCMessageBox(GetFocus(),"No se pudo abrir la salida de error simulada\n", "Error", MB_OK | MB_ICONEXCLAMATION);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Trap --
 *
 *	Atiende la trap que se solicita.  Los argumentos se esperan que
 *	comiencen en la direccion almacenada en r14 (el primer argumento)
 *	con otro argumentos en las palabras de memoria que le siguen. El
 *	resultado del la trap simulada es almacenado en r1.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio un error, 1 si la trap es
 *	conocida, pero hubo un problema con el manejador de la trap, o 2
 *	si el numero no es conocido.
 *
 * Efectos Laterales:
 *      El mapa de descriptores de ficheros de la memoria simulada del DLXV
 *	puede sufrir cambios.
 *
 *----------------------------------------------------------------------
 */
int Handle_Open();
int Handle_Close();
int Handle_Read();
int Handle_Write();
int Handle_Printf();

long int Handle_Trap(machPtr, trapNum)
DLXV *machPtr;
long int trapNum;
{
  switch (trapNum) {
  case TRAP_OPEN :
    return Handle_Open(machPtr);
  case TRAP_CLOSE :
    return Handle_Close(machPtr);
  case TRAP_READ :
    return Handle_Read(machPtr);
  case TRAP_WRITE :
    return Handle_Write(machPtr);
  case TRAP_PRINTF :
    return Handle_Printf(machPtr);
  case TRAP_PEND :
    return 3;
  }
  return 2;
}

/*
 *----------------------------------------------------------------------
 *
 * Block_To_Scratch --
 *
 *	Copia un bloque de la memoria simulada al scratch pad.
 *
 * Resultados:
 *	El valor devuelto es si no ocurrio un error, 1 si ocurrio algun
 *	problema.
 *
 * Efectos LAterales:
 *      El memScratch es modificado.
 *
 *----------------------------------------------------------------------
 */

long int Block_To_Scratch(machPtr, firstAddr, length)
DLXV *machPtr;
register long int firstAddr, length;
{
  char huge *first, huge *last, bytes[4];
  register unsigned long int word;
  MemWord huge *wordPtr;

  if (!length)
    return 0;
  length--;
  if (length < 0) {
    BWCCMessageBox(GetFocus(),"Block movement not Valid\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }
  if ((firstAddr < 0) || (firstAddr + length >= machPtr->numChars)) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }

  first = SCRATCH_ADDR(firstAddr);
  last = first + length;
  wordPtr = machPtr->memPtr + ADDR_TO_INDEX(firstAddr);
  word = wordPtr->value;
  bytes[3] = (word & 0xff);
  bytes[2] = ((word >>= 8) & 0xff);
  bytes[1] = ((word >>= 8) & 0xff);
  bytes[0] = ((word >> 8) & 0xff);
  while ((FP_SEG(first) < FP_SEG(last)) ||
  ((FP_SEG(first) == FP_SEG(last)) && (FP_OFF(first) <= FP_OFF(last)))) {
    *first = bytes[(unsigned int)(firstAddr & 0x3)];
    firstAddr++; first++;
    if (!(firstAddr & 0x3)) {
      word = (++wordPtr)->value;
      bytes[3] = (word & 0xff);
      bytes[2] = ((word >>= 8) & 0xff);
      bytes[1] = ((word >>= 8) & 0xff);
      bytes[0] = ((word >> 8) & 0xff);
    }
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Block_From_Scratch --
 *
 *	Copia un bloque de la memoria del scratch pad a la memoria
 *	simulada.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio ningun error, 1 si ocurrio
 *	algun problema.
 *
 * Efectos Laterales:
 *      El memPtr es modificado.
 *
 *----------------------------------------------------------------------
 */

long int Block_From_Scratch(machPtr, firstAddr, length)
DLXV *machPtr;
register long int firstAddr, length;
{
  char huge *first, huge *last;
  register unsigned long int word;
  MemWord huge *wordPtr;

  if (!length)
    return 0;
  length--;
  if (length < 0) {
    BWCCMessageBox(GetFocus(),"Block movement not Valid\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }
  if ((firstAddr < 0) || (firstAddr + length >= machPtr->numChars)) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }

  wordPtr = machPtr->memPtr + ADDR_TO_INDEX(firstAddr);
  word = wordPtr->value;
  first = SCRATCH_ADDR(firstAddr);
  last = first + length;
  while ((FP_SEG(first) < FP_SEG(last)) ||
  ((FP_SEG(first) == FP_SEG(last)) && (FP_OFF(first) <= FP_OFF(last))))  {
    switch (firstAddr & 0x3) {
    case 0 :
      word = (word & 0x00ffffffL) | ((long int)*first << 24);
      break;
    case 1 :
      word = (word & 0xff00ffffL) | ((long int)*first << 16);
      break;
    case 2 :
      word = (word & 0xffff00ffL) | ((long int)*first << 8);
      break;
    case 3 :
      word = (word & 0xffffff00L) | (long int)*first;
      wordPtr->value = word;
      wordPtr->opCode = OP_NOT_COMPILED;
      word = (++wordPtr)->value;
      break;
    }
    firstAddr++; first++;
  }
  if (wordPtr->value != word) {
    wordPtr->value = word;    /* Actualiza la ultima palabra */
    wordPtr->opCode = OP_NOT_COMPILED;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * String_To_Scratch --
 *
 *	Copia un string terminado con el caracter NULL de la memoria del
 *	simulador al scratch pad.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio un error, 1 si un algun
 *	problema ha aparecido.
 *
 * Efectos Laterales:
 *      El memScratch es modificado.
 *
 *----------------------------------------------------------------------
 */

long int String_To_Scratch(machPtr, firstAddr)
DLXV *machPtr;
register long int firstAddr;
{
  char huge *first,huge *endMem = machPtr->endScratch;
  register unsigned long int word;
  unsigned char bytes[4];
  MemWord huge *wordPtr;

  if ((firstAddr < 0) || (firstAddr >= machPtr->numChars)) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }

  first = SCRATCH_ADDR(firstAddr);
  wordPtr = machPtr->memPtr + ADDR_TO_INDEX(firstAddr);
  word = wordPtr->value;
  bytes[3] = (word & 0xff);
  bytes[2] = ((word >>= 8) & 0xff);
  bytes[1] = ((word >>= 8) & 0xff);
  bytes[0] = ((word >> 8) & 0xff);
  while ((FP_SEG(first) < FP_SEG(endMem)) ||
  ((FP_SEG(first) == FP_SEG(endMem)) && (FP_OFF(first) < FP_OFF(endMem)))) {
    if (!(*first = bytes[(unsigned int)(firstAddr & 0x3)]))
      break;
    firstAddr++; first++;
    if (!(firstAddr & 0x3)) {
      word = (++wordPtr)->value;
      bytes[3] = (word & 0xff);
      bytes[2] = ((word >>= 8) & 0xff);
      bytes[1] = ((word >>= 8) & 0xff);
      bytes[0] = ((word >> 8) & 0xff);
    }
  }

  if ((FP_SEG(first) > FP_SEG(endMem)) ||
  ((FP_SEG(first) == FP_SEG(endMem)) && (FP_OFF(first) >= FP_OFF(endMem)))) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n","Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Open --
 *
 *	Atiende la llamada al sistema open.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio un error, 1 si ocurrio algun
 *	problema.
 *
 * Efectos Laterales:
 *      El mapa de los descriptores de fichero puede cambiar.
 *
 *----------------------------------------------------------------------
 */
Set_Errno();

int Handle_Open(machPtr)
DLXV *machPtr;
{
  long int where, path, result;

  for (where = 0; where < FD_SIZE; where++)
    if (machPtr->fd_map[(int)where] == -1)
      break;
  if (where == FD_SIZE) {    /* no hay descriptores de fichero libres. */
    Set_Errno(machPtr, EMFILE);
    RETURN_VALUE = -1;
    return 0;
  }
  path = ARG(0);
  if (String_To_Scratch(machPtr, path))
    return 1;
  result = open((char far *)SCRATCH_ADDR(path),(int) ARG(1),(unsigned int) ARG(2));
  if (result == -1) {
    Set_Errno(machPtr, errno);
    RETURN_VALUE = -1;
  } else {
    machPtr->fd_map[(int)where] = result;
    RETURN_VALUE = where;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Close --
 *
 *	Maneja la llamada al sistema close.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio ningun error, 1 si existio
 *	algun problema.
 *
 * Efectos Laterales:
 *      El mapa de los descriptores de fichero puede cambiar.
 *
 *----------------------------------------------------------------------
 */

int Handle_Close(machPtr)
DLXV *machPtr;
{
  long int result, fd;

  fd = ARG(0);
  if ((fd < 0) || (fd > FD_SIZE) || (machPtr->fd_map[(int)fd] == -1)) {
    /* Descriptor de ficero ilegal */
    Set_Errno(machPtr, EBADF);
    RETURN_VALUE = -1;
    return 0;
  }
  result = close((int) machPtr->fd_map[(int)fd]);
  if (result == -1) {
    Set_Errno(machPtr, errno);
    RETURN_VALUE = -1;
  } else {
    RETURN_VALUE = result;
    machPtr->fd_map[(int)fd] = -1;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Read --
 *
 *	Maneja la llamada al sistema read.
 *
 * Resultados:
 *	El valor devuelto es 0 si no ocurrio ningun error, 1 si hubo algun
 *	problema.
 *
 * Efectos Laterales:
 *      La memoria del simulador y el scratch pad pueden cambiar.
 *
 *----------------------------------------------------------------------
 */

int Handle_Read(machPtr)
DLXV *machPtr;
{
  long int result, fd, buf, nbytes;

  fd = ARG(0);
  buf = ARG(1);
  nbytes = ARG(2);
  if ((fd < 0) || (fd > FD_SIZE) || (machPtr->fd_map[(int)fd] == -1)) {
    /* Descriptor de fichero ilegal */
    Set_Errno(machPtr, EBADF);
    RETURN_VALUE = -1;
    return 0;
  }
  if ((buf < 0) || (buf >= machPtr->numChars)) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }
  if ((nbytes < 0) || (buf + nbytes - 1 >= machPtr->numChars)) {
    BWCCMessageBox(GetFocus(),"Memory reference is out of range\n", "Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
  }
  result = read((int) machPtr->fd_map[(int)fd], SCRATCH_ADDR(buf),(unsigned int) nbytes);
  if (result == -1) {
    Set_Errno(machPtr, errno);
    RETURN_VALUE = -1;
  } else {
    if (Block_From_Scratch(machPtr, buf, result))
      return 1;
    RETURN_VALUE = result;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Write --
 *
 *	Maneja la llamada al sistema write.
 *
 * Resultados:
 *	El valor devuelto es 0 si no hubo problemas, 1 si ocurrio algun
 *	problema.
 *
 * Efectos Laterales:
 *      El scratch pad puede cambiar.
 *
 *----------------------------------------------------------------------
 */

int Handle_Write(machPtr)
DLXV *machPtr;
{
  long int result, fd, buf, nbytes;

  fd = ARG(0);
  buf = ARG(1);
  nbytes = ARG(2);
  if ((fd < 0) || (fd > FD_SIZE) || (machPtr->fd_map[(int)fd] == -1)) {
    /* Descriptor de fichero ilegal */
    Set_Errno(machPtr, EBADF);
    RETURN_VALUE = -1;
    return 0;
  }
  if (Block_To_Scratch(machPtr, buf, nbytes))
    return 1;
  result = write((int) machPtr->fd_map[(int)fd], SCRATCH_ADDR(buf),(unsigned int) nbytes);
  if (result == -1) {
    Set_Errno(machPtr, errno);
    RETURN_VALUE = -1;
  } else {
    RETURN_VALUE = result;
  }
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Handle_Printf --
 *
 *	Maneja la llamada al sistema printf.  Esto se realiza mirando el
 *	el string de formato proporcionado, y enviando cada comando % y su
 *	texto correspondiente a la funcion de libreria printf (con los
 *	argumentos apropiados).
 *
 * Resultados:
 *	El valor devuelto es cero si no ocurrio ningun error, 1 si ocurrio
 *	algun problema.
 *
 * Efectos Laterales:
 *      El scratch pad puede cambiar.
 *
 *----------------------------------------------------------------------
 */

int Handle_Printf(machPtr)
DLXV *machPtr;
{
  char huge *start, huge *end, huge *s, temp;
  long int fmt = ARG(0), argNumber = 1, ta[2], loc, done, result = 0;
  double *pd = (double *)&(ta[0]);

  /* Convierte el string de formato */
  if (String_To_Scratch(machPtr, fmt))
    return 1;

  /* Ahora separa el string de formato en bloques de caracteres empezando
   * con el principio del string de formato, o con el comando %, y
   * continuando cpn el principio del siguiente comando %, o con el final
   * del string. */

  for (start = SCRATCH_ADDR(fmt); *start; start = end) {
    if (*start == '%') {
      if (start[1] == '%')
	end = start + 2;
      else
	end = start + 1;
    } else
      end = start;
    for ( ; *end && (*end != '%'); end++)	/* Encuentra el final
						 * de este bloque */
      ;
    temp = *end;
    *end = '\0';
    if (*start != '%')
      result += printf((char far *)start);
    else {
      /* Encuentra el tipo de argumento que este bloque
       * toma ( si hay alguno). */
      for (done = 0, s = start + 1; *s && !done; s++)
	switch (*s) {
	case 'c' :
	case 'd' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' :
	  result += printf((char far *)start, ARG(argNumber++));
	  done = 1;
	  break;
	case 'f' :
	case 'e' :
	case 'E' :
	case 'g' :
	case 'G' :
	  ta[0] = ARG(argNumber++);
	  ta[1] = ARG(argNumber++);
	  result += printf((char far *)start, *pd);
	  done = 1;
	  break;
	case 's' :
	  loc = ARG(argNumber++);
	  if (String_To_Scratch(machPtr, loc))
	    return 1;
	  result += printf((char far *)start, SCRATCH_ADDR(loc));
	  done = 1;
	  break;
	case '%' :
	  result += printf((char far *)start);
	  done = 1;
	  break;
	}
      if (!done)    /* No se encontro el comando % */
	result += printf((char far *)start);
    }
    *end = temp;
  }
  RETURN_VALUE = result;
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Set_Errno
 *
 *	Asigna la variable _error (si existe)
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Ver la descripcion anterior.
 *
 *----------------------------------------------------------------------
 */

Set_Errno(machPtr, errorNumber)
DLXV *machPtr;
int errorNumber;
{
  char *p, *end;
  long int addr, index;

/* ***  p = Tcl_GetVar(machPtr->interp, "_errno", 0); *** */
/* *** */
  p="0x0";
  end="";
/* *** */
  if (*p != 0) {
    addr = strtoul(p, &end, 0);
    if (*end != 0)
      return;
    index = ADDR_TO_INDEX(addr);
    if (index < machPtr->numWords)
      machPtr->memPtr[index].value = errorNumber;
  }
  return;
}
