/*
 * Hash_CreateEntry.c --
 *
 *      Codigo fuente para el procedimiento de libreria Hash_CreateEntry.
 *
 * Copyright 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include <string.h>
#include "hash.h"
#include "list.h"

/*
 * Procedimientos utiles que estan definidos en otros ficheros:
 */

extern Hash_Entry *     HashChainSearch();
extern long int         Hash();

/*
 * Lo siguiente define la relacion entre # entradas y # buckets
 * con la cual reconstruiremos la tabla para hacerla mas grande.
 */

static rebuildLimit = 3;

/*
 *---------------------------------------------------------
 *
 * RebuildTable --
 *      Esta rutina local construye una nueva tabla hash que
 *      es mas grande que la antigua.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      La tabla en su totalidad es trasladada, por lo tanto todos
 *	los numeros de bucket de la vieja tabla son invalidos.
 *
 *---------------------------------------------------------
 */

static void
RebuildTable(tablePtr)
    Hash_Table  *tablePtr;              /* Tabla a ser aumentada. */
{
    long int             oldSize;
    long int             bucket;
    List_Links           *oldBucketPtr;
    register Hash_Entry  *hashEntryPtr;
    register List_Links  *oldHashList;

    oldBucketPtr = tablePtr->bucketPtr;
    oldSize = tablePtr->size;

    /*
     * Costruye una nueva tabla 4 veces mas grande que la antigua.
     */

    Hash_InitTable(tablePtr, tablePtr->size * 4, tablePtr->keyType);

    for (oldHashList = oldBucketPtr; oldSize > 0; oldSize--, oldHashList++) {
	while (!List_IsEmpty(oldHashList)) {
	    hashEntryPtr = (Hash_Entry *) List_First(oldHashList);
	    List_Remove((List_Links *) hashEntryPtr);
	    switch (tablePtr->keyType) {
		case HASH_STRING_KEYS:
		    bucket = Hash(tablePtr, (Address) hashEntryPtr->key.name);
		    break;
		case HASH_ONE_WORD_KEYS:
		    bucket = Hash(tablePtr, (Address) hashEntryPtr->key.ptr);
		    break;
		default:
		    bucket = Hash(tablePtr, (Address) hashEntryPtr->key.words);
		    break;
	    }
	    List_Insert((List_Links *) hashEntryPtr,
		LIST_ATFRONT(&(tablePtr->bucketPtr[(unsigned int)bucket])));
	    tablePtr->numEntries++;
	}
    }

    farfree((Address) oldBucketPtr);
}

/*
 *---------------------------------------------------------
 *
 * Hash_CreateEntry --
 *
 *      Busca en la tbla hash una entrada correspondiente a una clave
 *      (key).  Si no es encontrada ninguna entrada, entonces se crea una.
 *
 * Resultados:
 *      El valor devuelto es un puntero a la entrada.  Si *newPtr
 *      no es NULL, entonces en *newPtr se almacena el valor TRUE si una
 *      nueva entrada fue creada, y FALSE si existe una entrada con la
 *      clave que se proporciona.
 *
 * Efectos Laterales:
 *      La memoria puede ser alojada, y los buckets pueden ser modificados.
 *---------------------------------------------------------
 */

Hash_Entry *
Hash_CreateEntry(tablePtr, key, newPtr)
    register Hash_Table *tablePtr;      /* Tabla hash en la que buscar. */
    Address key;                        /* La clave hash. */
    Boolead *newPtr;                    /* Rellenada con TRUE si se crea nueva
					 * entrada, FALSE en cualquier otro
					 * caso. */
{
    register Hash_Entry *hashEntryPtr;
    register long int   *hashKeyPtr;
    register long int   *keyPtr;
    List_Links          *hashList;

    keyPtr = (long int *) key;

    hashList = &(tablePtr->bucketPtr[(unsigned int)Hash(tablePtr, (Address) keyPtr)]);
    hashEntryPtr = HashChainSearch(tablePtr, (Address) keyPtr, hashList);

    if (hashEntryPtr != (Hash_Entry *) NULL) {
	if (newPtr != NULL) {
	    *newPtr = FALSE;
	}
	return hashEntryPtr;
    }

    /* 
     * La entrada deseada no esta aqui.  Antes de alojar una nueva entrada,
     * vemos si estamos sobrecargando los buckets.  Si es asi, entonces
     * construimos una tabla mayor.
     */

    if (tablePtr->numEntries >= rebuildLimit * tablePtr->size) {
	RebuildTable(tablePtr);
	hashList = &(tablePtr->bucketPtr[(unsigned int)Hash(tablePtr, (Address) keyPtr)]);
    }
    tablePtr->numEntries += 1;

    /*
     * No estaba, entonces tenemos que alojar.  Si el string es mayor de
     * 3 bytes, tenemos que alojar espacio extra en la entrada.
     */

    switch (tablePtr->keyType) {
	case HASH_STRING_KEYS:
	    hashEntryPtr = (Hash_Entry far *) farcalloc(1, (unsigned long int)
		    (sizeof(Hash_Entry) + _fstrlen((char far *) keyPtr) - 3));
	    strcpy(hashEntryPtr->key.name, (char far *) keyPtr);
	    break;
	case HASH_ONE_WORD_KEYS:
	    hashEntryPtr = (Hash_Entry far *) farcalloc(1, sizeof(Hash_Entry));
	    hashEntryPtr->key.ptr = (Address) keyPtr;
	    break;
	case 2:
	    hashEntryPtr =
		(Hash_Entry far *) farcalloc(1, sizeof(Hash_Entry) + sizeof(long int));
	    hashKeyPtr = hashEntryPtr->key.words;
	    *hashKeyPtr++ = *keyPtr++;
	    *hashKeyPtr = *keyPtr;
	    break;
	default: {
	    register long int n;

	    n = tablePtr->keyType;
	    hashEntryPtr = (Hash_Entry far *)
		    farcalloc(1, (unsigned long int) (sizeof(Hash_Entry)
		    + (n - 1) * sizeof(long int)));
	    hashKeyPtr = hashEntryPtr->key.words;
	    do {
		*hashKeyPtr++ = *keyPtr++;
	    } while (--n);
	    break;
	}
    }

    hashEntryPtr->clientData = (ClientData) NULL;
    List_Insert((List_Links *) hashEntryPtr, LIST_ATFRONT(hashList));

    if (newPtr != NULL) {
	*newPtr = TRUE;
    }
    return hashEntryPtr;
}
