/*
 * Hash_EnumNext.c --
 *
 *	Codigo fuente para el procedimiento de libreria Hash_EnumNext.
 *
 * Copyright 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include "hash.h"
#include "list.h"

/*
 *---------------------------------------------------------
 *
 * Hash_EnumNext --
 *    Este procedimiento retorna sucesivas entradas de la tabla hash.
 *
 * Resultados:
 *    El valor de retorno es un puntero a la siguiente entrada (HashEntry)
 *    de la tabla, o NULL cuando se llega al final de la tabla.
 *
 * Efectos Laterales:
 *    La informacion almacenada en hashSearchPtr es modificada para avanzar
 *    a la siguiente entrada.
 *
 *---------------------------------------------------------
 */

Hash_Entry *
Hash_EnumNext(hashSearchPtr)
    register Hash_Search *hashSearchPtr; /* Area usda para mantener el
					    estado de la busqueda. */
{
    register List_Links *hashList;
    register Hash_Entry *hashEntryPtr;

    hashEntryPtr = hashSearchPtr->hashEntryPtr;
    while (hashEntryPtr == (Hash_Entry *) NULL ||
	   List_IsAtEnd(hashSearchPtr->hashList,
	   (List_Links *) hashEntryPtr)) {
	if (hashSearchPtr->nextIndex >= hashSearchPtr->tablePtr->size) {
	    return((Hash_Entry *) NULL);
	}
	hashList = &(hashSearchPtr->tablePtr->bucketPtr[
		(unsigned int)hashSearchPtr->nextIndex]);
	hashSearchPtr->nextIndex++;
	if (!List_IsEmpty(hashList)) {
	    hashEntryPtr = (Hash_Entry *) List_First(hashList);
	    hashSearchPtr->hashList = hashList;
	    break;
	}
    }

    hashSearchPtr->hashEntryPtr =
		(Hash_Entry *) List_Next((List_Links *) hashEntryPtr);

    return(hashEntryPtr);
}
