/*
 * tcl.h --
 *
 *	Este fichero de cabecera describe las facilidades del interprete
 *	Tcl que pueden ser vistas desde el exterior. Solo unas pocas
 *	de estas facilidades han sido utilizadas en la realizacion
 *	de este proyecto.
 *
 * Copyright 1987 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef _TCL
#define _TCL


/*
 * Cuando un comando termina, la variable machPtr->result apunta a una
 * cadnea que contiene la informacion devuelta por el comando. Ademas el
 * coando devuelve un valor entero, el cual es uno de los siguientes:
 *
 * TCL_OK		El comando termino normalmente;  machPtr->result
 *			contiene el resultado del comando.
 * TCL_ERROR		El comando no pudo ser completado satisfactoriamente;
 *			machPtr->result describe el problema ocurrido.
 * TCL_RETURN		El comando pide que el actual procedimiento devuelva
 *			un valor; machPtr->result contiene el valor devuelto
 *			por el procedimiento.
 * TCL_BREAK		El comando pide que se salga del bucle en el que se
 *			encuentre; machPtr->result no tiene ningun significado.
 * TCL_CONTINUE		Continua con la siguiente iteracion del bucle actual;
 *			machPtr->resul no tiene ningun significado.
 */

#define TCL_OK		0L
#define TCL_ERROR	1L
#define TCL_RETURN	2L
#define TCL_BREAK	3L
#define TCL_CONTINUE	4L

#define TCL_RESULT_SIZE 60L

/*
 * Flags que se pasan a Tcl_Return:
 */

#define TCL_STATIC	0L
#define TCL_DYNAMIC	1L
#define TCL_VOLATILE	2L

/*
 * Procedimientos exportados por el Tcl:
 */

extern char		Tcl_Backslash();
extern void		Tcl_Return();

/*
 * Declaraciones varias (Para permitir que el Tcl pueda ser usado
 * de forma individual, sin el resto del Sprite):
 */

#ifndef NULL
#define NULL 0L
#endif

#ifndef _CLIENTDATA
typedef long int *ClientData;
#define _CLIENTDATA
#endif

#endif _TCL
